#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
Data objects for music.
'''

# pylint: disable=too-many-instance-attributes

from musicscan.data.flags import Flags
from musicscan.data.titletools import ShortTitleIndex, ShortTitle
from musicscan.data.track import Track
from musicscan.data.stringtools import build_complete_filename


class Album():
    '''
    Object class album representing an actual purchased
    CD, which could include one or more discs, with each
    disc containing one or more tracks.
    '''
    def __init__(self, in_artist, in_album_title):
        self.title = in_album_title
        self.artist = in_artist
        self.title_index = ShortTitleIndex()
        self.discs = {}
        self.flags = Flags()
        self.track_count = 0

    def import_tag(self, in_tag):
        '''
        Import an ID3 tag into the album, creating
        a Disc object if necessary.
        '''
        if not in_tag.disc:
            disc_no = 1
        else:
            disc_no = int(in_tag.disc)
        if disc_no not in self.discs:
            self.discs[disc_no] = Disc(disc_no, self)
            self.discs[disc_no].import_tag(in_tag, self)
        else:
            self.discs[disc_no].import_tag(in_tag, self)
        self.track_count += 1

    def report(self):
        '''
        Generate a simple report on the ablum.
        '''
        print(f"Album: {self.title}")
        print(f"Artist: {self.artist}")
        for dsc in sorted(self.discs):
            self.discs[dsc].report()

    def summary(self):
        '''
        Generate a summary of what the album contains.
        '''
        output = ''
        first = self.first_track()
        output += f"Album: {self.title} ({first.year})\n"
        output += f"Artist: {self.artist!s}\n"
        output += f"Discs: {len(self.discs)}  Tracks: {self.track_count}\n"
        return output

    def filename(self, in_suffix='album'):
        '''
        Create a filename based on the interpreted ID3 album name.
        '''
        return build_complete_filename(self, in_suffix)

    def first_track(self):
        '''
        Return the very first track of the album.
        '''
        return self.discs[1].tracks[0]

    def tracks_in_order(self):
        '''
        Return all tracks from all discs on the album.
        '''
        all_tracks = []
        for dsc in sorted(self.discs):
            for trk in sorted(self.discs[dsc].tracks,
                              key=lambda x: x.track_no):
                all_tracks.append(trk)
        return all_tracks

    def finalize(self):
        '''
        Finalize each object by setting up the short titles
        for each track.
        '''
        for dsc in sorted(self.discs):
            self.discs[dsc].finalize()

    def __str__(self):
        return f"{self.title} ({len(self.discs)})"


class Disc():
    '''
    Class representing a single physical compact disc.
    '''
    def __init__(self, in_number, in_album):
        self.disc_no = in_number
        self.album = in_album
        self.tracks = []
        self.raw = []

    def import_tag(self, in_tag, in_album):
        '''
        Import data from an ID3 tag into the disc object.
        '''
        self.raw.append(in_tag)
        trk = Track(in_tag)
        trk.set_album_object(in_album)
        self.tracks.append(trk)

    def finalize(self):
        '''
        Update data after all import operations are done.
        '''
        for trk in sorted(self.tracks):
            trk.short_title = ShortTitle(trk.title, self.album.title_index)

    def report(self):
        '''
        Generate a simple report on the disc object.
        '''
        print(f"Disc: {self.disc_no}")
        for trk in sorted(self.tracks, key=lambda x: x.track_no):
            print(f"{trk}")

    def __lt__(self, other):
        return self.disc_no < other.disc_no

    def __gt__(self, other):
        return self.disc_no > other.disc_no



#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
Code to scan one or more directories for media files
'''

# media.scanner

import os.path
import time
from os import listdir
from os.path import isdir, isfile


class Walker():
    '''Scans directory for media XML files'''
    def __init__(self, in_paths, debug=False):
        self.dirs = []
        self.files = []
        self.skipped = []
        self.elapsed = None
        self.match_re = None
        self.debug = debug
        self.dirs.append(in_paths[0])

    def filename_match(self, in_match):
        '''Pass filename matching regex'''
        self.match_re = in_match

    def scan(self):
        '''Scan the directory tree for files'''
        tstart = time.time()
        for s_dir in self.dirs:
            for s_file in listdir(s_dir):
                full_path = os.path.join(s_dir, s_file)
                if isdir(full_path):
                    if not s_file.startswith('.'):
                        self.dirs.append(full_path)
                elif isfile(full_path):
                    self.files.append(full_path)
        tend = time.time()
        self.elapsed = tend - tstart

    def name_match(self, in_filename):
        '''Test a filename to ensure it matches our search pattern'''
        if self.match_re:
            if self.match_re.search(in_filename):
                return True
        return False

    def stats(self):
        '''Simple statistical output'''
        return f"{len(self.dirs)} {len(self.files)} {len(self.skipped)}"

    def pass_stats(self, stats):
        '''
        Update the stats object with information on
        file scanning operation.
        '''
        stats.source_dirs = len(self.dirs)
        stats.source_files = len(self.files)
        stats.source_skipped = len(self.skipped)

    def report(self):
        '''More detailed report of scanner operations'''
        out = "Stats\n-----\n"
        out += f"Directory Count   : {len(self.dirs)}\n"
        out += f"File Count        : {len(self.files) + len(self.skipped)}\n"
        out += f"        Matched   : {len(self.files)}\n"
        out += f"    Non-Matched   : {len(self.skipped)}\n"
        out += f"Elapsed Time      : {self.elapsed:02f}\n"
        return out



#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
Testing code to identify possible edit flags.
'''

import re
from musicscan.data.flags import FlagCodes


class Analyzer():
    '''
    Perform tests against the data and
    add flags when appropriate.
    '''

    def main_test(self, in_album):
        '''
        Main album testing routine.
        '''
        self.test_album_title(in_album)
        self.test_album_artist(in_album)
        self.test_soundtrack(in_album)
        self.test_score(in_album)
        for dsc in sorted(in_album.discs):
            for trk in sorted(in_album.discs[dsc].tracks,
                              key=lambda x: x.track_no):
                self.test_album_track_chrs(trk)
                self.test_album_track_featured(trk)
                self.test_album_track_live(trk)
                self.test_album_track_blank(trk)

    def test_album_title(self, in_album):
        '''
        Test album title for keywords that may indicate
        a compilation of some kind.
        '''
        greatest_p = re.compile(r'Greatest', re.IGNORECASE)
        hits_p = re.compile(r'\s+Hit', re.IGNORECASE)
        if greatest_p.search(in_album.title):
            in_album.flags.add_flag(FlagCodes.p_greatest)
        if hits_p.search(in_album.title):
            in_album.flags.add_flag(FlagCodes.p_hits_compo)

    def test_soundtrack(self, in_album):
        '''
        Test album title for keywords that may indicate
        a soundtrack album.
        '''
        first_track = in_album.first_track()
        soundtrack_p = re.compile(r'\s+Soundtrack\s+', re.IGNORECASE)
        if soundtrack_p.search(in_album.title):
            in_album.flags.add_flag(FlagCodes.p_soundtrack)
        if first_track.genre == 'Soundtrack':
            if FlagCodes.p_soundtrack not in in_album.flags.flags:
                in_album.flags.add_flag(FlagCodes.p_soundtrack)

    def test_score(self, in_album):
        '''
        Test album title for keywords that may indicate
        a score album.
        '''
        score_p = re.compile(r'\s+Score\s+', re.IGNORECASE)
        if score_p.search(in_album.title):
            in_album.flags.add(FlagCodes.p_score)

    def test_album_artist(self, in_album):
        '''
        Test the name of the artist for indicators
        the band may be a group.
        '''
        and_p = re.compile(r'\s+and\s+', re.IGNORECASE)
        if and_p.search(str(in_album.artist)):
            in_album.flags.add_flag(FlagCodes.l_group)

    def test_album_track_chrs(self, in_track):
        '''
        Test the track title to indicate metadata may
        have been embeded in the track name in parenthesis.
        '''
        if '[' in in_track.title or ']' in in_track.title:
            in_track.flags.add_flag(FlagCodes.d_sq_brackets)
        if '(' in in_track.title or ')' in in_track.title:
            in_track.flags.add_flag(FlagCodes.d_parenthesis)

    def test_album_track_featured(self, in_track):
        '''
        Test the track title or track artist to see if
        there may be additional artists for this particular track.
        '''
        feature_p = re.compile(r'Feat', re.IGNORECASE)
        with_p = re.compile(r'With\s+', re.IGNORECASE)
        if feature_p.search(in_track.title) or with_p.search(in_track.title):
            in_track.flags.add_flag(FlagCodes.p_feat_artist)
        if feature_p.search(in_track.artist) or with_p.search(in_track.artist):
            in_track.flags.add_flag(FlagCodes.p_feat_artist)

    def test_album_track_live(self, in_track):
        '''
        Test the track title to see if it might
        indicate a live recording.
        '''
        live_p = re.compile(r'Live\s+', re.IGNORECASE)
        if live_p.search(in_track.title):
            in_track.flags.add_flag(FlagCodes.p_live)

    def test_track_genre(self, in_track):
        '''
        Test the track for bad genre values.
        '''
        country_p = re.compile(r'Country & Folk', re.IGNORECASE)
        if country_p.search(in_track.genre):
            in_track.flags.add_flag(FlagCodes.p_genre_country_folk)

    def test_album_track_demo(self, in_track):
        '''
        Test the track title to see if it might
        be a demo recording.
        '''
        demo_p = re.compile(r'Demo', re.IGNORECASE)
        if demo_p.search(in_track.title):
            in_track.flags.add_flag(FlagCodes.p_demo)

    def test_album_track_blank(self, in_track):
        '''
        Test the track title to see if it might
        be a blank track to pad the CD content.
        '''
        blank_p = re.compile(r'Blank', re.IGNORECASE)
        if blank_p.search(in_track.title):
            in_track.flags.add_flag(FlagCodes.p_blank_track)



#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
Write out XML data for an album.
'''

import os.path
from musicscan.xml.builder import (CompleteCompactDiscXML,
                                   SplitCompactDiscXML, Index, Album)


class XMLFileWriter():
    '''
    Class to write out XML files.

    Handles opening files, writing data, and closing the handle.
    '''
    def __init__(self, in_path=None):
        self._path = in_path
        self._debug = False
        self._overwrite = False
        self._split_xml = False
        self.files_written = 0

    def debug(self):
        '''
        Return the debug value.
        '''
        return self._debug

    def set_debug(self, in_debug):
        '''
        Set the debug value.
        '''
        self._debug = in_debug

    def overwrite(self):
        '''
        Return the overwrite value.
        '''
        return self._overwrite

    def set_overwrite(self, in_overwrite):
        '''
        Set the overwrite value.
        '''
        self._overwrite = in_overwrite

    def split_xml(self):
        '''
        Return the split_xml value.
        '''
        return self._split_xml

    def set_split_xml(self, in_split_xml):
        '''
        Set the split_xml value.
        '''
        self._split_xml = in_split_xml

    def identify_files(self, in_album, in_type='audiocd'):
        '''
        Identify all possible files from an passed album.
        '''
        file_list = []
        file_list.append(self._path + '/' + in_album.filename(in_type))
        if self._split_xml:
            file_list.append(self._path + '/' + in_album.filename('album'))
            for dsc in sorted(in_album.discs):
                dsc_o = in_album.discs[dsc]
                f_str = f"cd{dsc_o.disc_no:02}-index"
                file_list.append(self._path + '/' + in_album.filename(f_str))
        return file_list

    def write_xml(self, in_album):
        '''
        Write out the XML data.
        '''
        if self._split_xml:
            self.write_split_xml(in_album)
        else:
            self.write_single_xml(in_album)

    def write_single_xml(self, in_album):
        '''
        Write all the XML data in a single file.
        '''
        complete = self._path + '/' + in_album.filename('audiocd')
        audiocd = CompleteCompactDiscXML()
        if not os.path.isfile(complete) or self._overwrite:
            with open(complete, mode='w', encoding='utf-8') as f_handle:
                f_handle.write(audiocd.build(in_album))
                f_handle.close()
                self.files_written += 1

    def write_split_xml(self, in_album):
        '''
        Write multiple XML files.
        '''
        self.write_split_xml_audiocd(in_album)
        self.write_split_xml_index(in_album)
        self.write_split_xml_album(in_album)

    def write_split_xml_audiocd(self, in_album):
        '''
        Write the XML data pertaining to the audio CD.
        '''
        complete = self._path + '/' + in_album.filename('audiocd')
        audiocd = SplitCompactDiscXML()
        if not os.path.isfile(complete) or self._overwrite:
            with open(complete, mode='w', encoding='utf-8') as f_handle:
                f_handle.write(audiocd.build(in_album))
                f_handle.close()
                self.files_written += 1

    def write_split_xml_index(self, in_album):
        '''
        Write the XML data pertaining to the CD index.
        '''
        index = Index()
        for dsc in sorted(in_album.discs):
            dsc_o = in_album.discs[dsc]
            f_str = f"cd{dsc_o.disc_no:02}-index"
            complete = self._path + '/' + in_album.filename(f_str)
            if not os.path.isfile(complete) or self._overwrite:
                with open(complete, mode='w', encoding='utf-8') as f_handle:
                    f_handle.write(index.build_index_per_cd(dsc_o, True))
                    f_handle.close()
                    self.files_written += 1

    def write_split_xml_album(self, in_album):
        '''
        Write out the XML data pertaining to the album contents.
        '''
        complete = self._path + '/' + in_album.filename('album')
        album = Album()
        if not os.path.isfile(complete) or self._overwrite:
            with open(complete, mode='w', encoding='utf-8') as f_handle:
                f_handle.write(album.build_standalone_album(in_album))
                f_handle.close()
                self.files_written += 1



#!/usr/bin/env python

#
# Copyright 2023 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
Object for collecitng basic statistics.
'''

from datetime import datetime


class Stats():
    '''
    Statistics from all operations of a given tool.
    '''
    def __init__(self):
        self.start_time = datetime.now()
        self.end_time = None
        self.source_dirs = 0
        self.source_files = 0
        self.album_count = 0
        self.track_count = 0
        self.files_written = 0

    def close(self):
        '''
        Save the end time for the program run.
        '''
        self.end_time = datetime.now()

    def report(self):
        '''
        Generate a simple report string.
        '''
        out = ''
        out += "\nTimes\n=====\n"
        out += f"{'Start Time':20s}: {self.start_time}\n"
        out += f"{'End Time':20s}: {self.end_time}\n"
        out += f"{'Run Duration':20s}: {self.end_time - self.start_time}\n"
        out += "\nInput Files\n===========\n"
        out += f"{'Directories Scanned':20s}: {self.source_dirs}\n"
        out += f"{'Files Scanned':20s}: {self.source_files}\n"
        out += "\nMusic Breakdown\n===============\n"
        out += f"{'Album Count':20s}: {self.album_count}\n"
        out += f"{'Total Tracks':20s}: {self.track_count}\n"
        out += "\nOutput Files\n============\n"
        out += f"{'Files Written':20s}: {self.files_written}\n"
        out += "\n"
        return out



#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


'''
XML Output blocks
'''

import datetime
from musicscan.data.stringtools import sanitize_for_xml


XML_DECLARATION = "<?xml version='1.0'?>\n"


class AbstractCompactDiscXML():
    '''
    Common routines written out during the creation of the
    main medium file for the CD.
    '''
    def __init__(self, in_debug=False):
        self.debug = in_debug

    def build_head(self):
        '''
        Build the start of the album XML.
        '''
        timestamp = datetime.datetime.now()
        output = XML_DECLARATION
        output += "<medialist xmlns='http://vectortron.com/xml/media/media'"
        output += " xmlns:xi='http://www.w3.org/2001/XInclude'>\n"
        output += f" <!-- created by id3scan ({timestamp}) -->\n"
        output += " <media>\n"
        return output

    def build_foot(self):
        '''
        XML Footer
        '''
        output = " </media>\n" +\
                 "</medialist>\n"
        return output

    def build_title(self, in_title):
        '''
        XML Title Block
        '''
        output = "  <title>\n"
        output += f"   <main>{in_title}</main>\n"
        output += " </title>\n"
        return output

    def build_medium(self, in_album):
        '''
        The main elements reporting on the physical media.
        '''
        output = "  <medium>\n"
        output += "  <release>\n"
        output += "   <type><audiocd/></type>\n"
        output += "  </release>\n"
        output += "  <productSpecs>\n"
        output += "   <inventory>\n"
        output += "    <case>\n"
        for dsc in sorted(in_album.discs):
            output += f"     <cd id='cd{in_album.discs[dsc].disc_no:02}'/>\n"
        output += "    </case>\n"
        output += "   </inventory>\n"
        output += "  </productSpecs>\n"
        output += " </medium>\n"
        return output


class CompleteCompactDiscXML(AbstractCompactDiscXML):
    '''
    Builds an XML structure in a single file.
    '''
    def build(self, in_album):
        '''
        Return the CD structure as a string.
        '''
        output = self.build_head()
        output += self.build_title(in_album.title)
        output += self.build_medium(in_album)
        index = Index()
        output += index.build_single_chunk_from_album(in_album)
        output += " <contents>\n"
        album = Album()
        output += album.build_album_body(in_album)
        output += " </contents>\n"
        output += self.build_foot()
        return output


class SplitCompactDiscXML(AbstractCompactDiscXML):
    '''
    Build an XML structure in multiple files.
    '''
    def build(self, in_album):
        '''
        Return the CD structure as a string.
        '''
        output = self.build_head()
        output += self.build_title(in_album.title)
        output += self.build_medium(in_album)
        index = Index()
        output += index.build_xi_chunk_from_album(in_album)
        output += " <contents>\n"
        album = Album()
        output += album.build_album_xi(in_album)
        output += " </contents>\n"
        output += self.build_foot()
        return output


class Index():
    '''
    Index element, which contains the track/index numbers for the content.
    '''
    def __init__(self, in_debug=False):
        self.debug = in_debug

    def build_single_chunk_from_album(self, in_album):
        '''
        Build an entire index element block with child elements.
        '''
        output = ''
        output += " <index>\n"
        output += self.build_multiple_chunks_from_album(in_album)
        output += " </index>\n"
        return output

    def build_xi_chunk_from_album(self, in_album):
        '''
        Build an index element block with xi:include elements.
        '''
        output = " <index>\n"
        output += self.build_xi_refs_from_album(in_album)
        output += " </index>\n"
        return output

    def build_multiple_chunks_from_album(self, in_album):
        '''
        String together multiple disc index strings together.
        '''
        chunks = []
        for dsc in sorted(in_album.discs):
            d_str = self.build_index_per_cd(in_album.discs[dsc])
            chunks.append(d_str)
        return "".join(chunks)

    def build_index_per_cd(self, in_cd, in_namespace=False):
        '''
        Build an index structure for a single CD.
        '''
        ns_str = ''
        if in_namespace:
            ns_str = " xmlns='http://vectortron.com/xml/media/media'"
        output = f"  <cdIndex ref='cd{in_cd.disc_no:02}'{ns_str}>\n"
        for trk in sorted(in_cd.tracks):
            output += self.add_track_xml(trk)
        output += "  </cdIndex>\n"
        return output

    def build_xi_refs_from_album(self, in_album):
        '''
        Output all of the xi:include references in the index.
        '''
        chunks = []
        for dsc in sorted(in_album.discs):
            d_str = self.build_xi_per_cd(in_album.discs[dsc])
            chunks.append(d_str)
        return "".join(chunks)

    def build_xi_per_cd(self, in_cd):
        '''
        Return an XI Include reference for every CD
        in the set.
        '''
        alb = in_cd.album
        f_str = f"cd{in_cd.disc_no:02}-index"
        filename = alb.filename(f_str)
        output = f"  <xi:include href='{filename}'/>\n"
        return output

    def add_track_xml(self, in_track):
        '''
        Output the information for a single track/song.
        '''
        output = ''
        output += f"   <track no='{in_track.track_no}'>\n"
        output += "    <index no='01'>\n"
        output += f"     <content ref='{in_track.short_title}'/>"
        if self.debug:
            output += f"   <!-- TITLE {in_track.title} -->"
        output += "\n    </index>\n   </track>\n"
        return output


class Album():
    '''
    XML Element structure representing an Album.
    '''
    def __init__(self, in_debug=False):
        self.debug = in_debug

    def build_standalone_album(self, in_album):
        '''
        Write out an entire album in a separate file.
        '''
        output = XML_DECLARATION
        output += self.build_album_body(in_album)
        return output

    def build_album_body(self, in_album):
        '''
        Write out the main body of the album element.
        '''
        output = "  <album xmlns='http://vectortron.com/xml/media/audio'>\n"
        output += f"   <title>{in_album.title}</title>\n"
        output += in_album.flags.to_xml_comment()
        output += self.build_chunk_catalog(in_album)
        output += self.build_chunk_classification(in_album)
        output += self.build_chunk_elements(in_album)
        output += "  </album>\n"
        return output

    def build_album_xi(self, in_album):
        '''
        Write out an xi:include element referencing the album.
        '''
        output = ''
        f_str = in_album.filename('album')
        output = f"  <xi:include href='{f_str}'/>\n"
        return output

    def build_chunk_catalog(self, in_album):
        '''
        Write out the album catalog element.
        '''
        first_track = in_album.first_track()
        output = " <catalog>\n"
        output += "  <artists>\n"
        output += f"   {in_album.artist.to_xml()}\n"
        output += "  </artists>\n"
        if first_track.year != '0000':
            output += "  <copyright>\n" +\
                      f"   <year>{first_track.year}</year>\n" +\
                      "  </copyright>\n"
        output += " </catalog>\n"
        return output

    def build_chunk_classification(self, in_album):
        '''
        Write out the album classification element.
        '''
        first_track = in_album.first_track()
        output = " <classification>\n"
        output += "  <genres>\n"
        output += f"   <primary>{sanitize_for_xml(first_track.genre)}"
        output += "</primary>\n  </genres>\n"
        output += " </classification>\n"
        return output

    def build_chunk_elements(self, in_album):
        '''
        Output the element element and all
        child content elements.
        '''
        output = " <elements>\n"
        song = SongElementXML(self.debug)
        for dsc in sorted(in_album.discs):
            for trk in in_album.discs[dsc].tracks:
                output += song.build(trk)
        output += " </elements>\n"
        return output


class SongElementXML():
    '''
    XML output covering a single element.

    We assume every piece of element content is a song
    because it's a safe bet, and there's no metadata
    that can tell us otherwise.
    '''
    def __init__(self, in_debug=False):
        self.debug = in_debug

    def build(self, in_track):
        '''
        Output the basic song element.
        '''
        output = f"  <song id='{in_track.short_title}'>\n"
        if self.debug:
            output += f"  <!-- POSITION: Disc {in_track.disc_no} " +\
                      f"Track: {in_track.track_no} -->\n"
        output += "   <title>\n"
        output += f"    <main>{sanitize_for_xml(in_track.title)}</main>\n"
        output += "  </title>\n"
        output += in_track.flags.to_xml_comment()
        output += self.build_catalog(in_track)
        output += self.build_technical(in_track)
        output += " </song>\n"
        return output

    def build_catalog(self, in_track):
        '''
        Output the per track catalog element (if it exists)
        '''
        # output += in_track.to_xml_catalog()
        output = ''
        if in_track.artist != str(in_track.album_o.artist) \
                or in_track.composer:
            output = "  <catalog>\n"
            if in_track.artist != str(in_track.album_o.artist):
                output += "   <artists>\n" +\
                          "    <artist><unkn>" +\
                          f"{sanitize_for_xml(in_track.artist)}" +\
                          "</unkn></artist>\n" +\
                          "   </artists>\n"
            if in_track.composer:
                output += "   <composers>\n" +\
                          "    <composer><unkn>" +\
                          f"{sanitize_for_xml(in_track.composer)}" +\
                          "</unkn></composer>\n" +\
                          "   </composers>\n"
            output += "   </catalog>\n"
        return output

    def build_technical(self, in_track):
        '''
        Output the XML technical element for the song.
        '''
        output = "   <technical>\n"
        output += "    <studioRecording/>\n"
        output += "    <runtime>\n"
        output += f"      <overall>{in_track.duration_s}</overall>\n"
        output += "    </runtime>\n"
        if in_track.bpm > 0:
            output += "    <tempo>\n"
            output += f"     <bpm>{in_track.bpm}</bpm>\n"
            output += "   </tempo>\n"
        output += "  </technical>\n"
        return output



#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
id3scam

Scan a directory full of music files for ID3 data,
parse the data, and then generate XML files for
each musical album.

(Assuming each album is a musical CD)

'''

import argparse
import os
import os.path
import sys
from tinytag import TinyTag
from musicscan.fileops.scanner import Walker
from musicscan.fileops.filenames import FilenameMatches
from musicscan.fileops.xmlwriter import XMLFileWriter
from musicscan.generic.stats import Stats
from musicscan.data.library import Library, Organizer


def get_files(in_path, in_stats):
    '''
    Get all the music files from the stated path.
    '''
    file_list = []
    walker = Walker([in_path], debug=True)
    walker.scan()
    walker.pass_stats(in_stats)
    pattern = FilenameMatches.Music_Media
    for file in walker.files:
        if pattern.search(file):
            file_list.append(file)
    return file_list


def scan_files(in_files):
    '''
    Generate a TinyTag object from each file identified.
    '''
    tag_objects = []
    for file in in_files:
        print(f"Source file : {file}")
        tagdata = TinyTag.get(file)
        tag_objects.append(tagdata)
    return tag_objects


def report_xml_files(in_args, in_organizer):
    '''
    Report on the XML files that will be generated
    from the collected data.
    '''
    done = []
    not_done = []
    writer = XMLFileWriter(in_args.outdir)
    writer.set_split_xml(in_args.splitxml)
    for alb_i in in_organizer.albums:
        out_files = writer.identify_files(alb_i)
        for file in out_files:
            if os.path.isfile(file):
                done.append(file)
            else:
                not_done.append(file)
    print("\nExisting XML Files\n------------------\n")
    for alxml in done:
        print(alxml)
    print("\nNon-existing XML files\n--------------\n")
    for alxml in not_done:
        print(alxml)


def write_xml_files(in_organizer, in_args, in_stats):
    '''
    Write out the XML files with the accumulated ID3 data.
    '''
    if not os.path.isdir(in_args.outdir):
        print("\nFAILURE: Cannot write to output path {in_args.outdir}\n")
        sys.exit(1)
    writer = XMLFileWriter(in_args.outdir)
    writer.set_debug(in_args.debug)
    writer.set_overwrite(in_args.overwrite)
    writer.set_split_xml(in_args.splitxml)
    for alb_i in in_organizer.albums:
        writer.write_xml(alb_i)
    in_stats.files_written = writer.files_written


def show_debug_data(in_data):
    '''
    Output a bunch of data for debugging purposes.
    '''
    for tag_entry in in_data:
        print(f"\nTitle: {tag_entry.title} {tag_entry.track}\n")
        entry_d = tag_entry.as_dict()
        for field in entry_d:
            print(f"Field: {field} {entry_d[field]}")


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Scan music files',
                                     epilog='Report music data and output XML')
    parser.add_argument('--musicpath', help='path of music files')
    parser.add_argument('--write', action='store_true',
                        default=False,
                        help='write out XML files')
    parser.add_argument('--split-xml', action='store_true',
                        default=False,
                        dest='splitxml',
                        help='split output XML files')
    parser.add_argument('--outdir', help='XML output directory')
    parser.add_argument('--overwrite', action='store_true',
                        default=False,
                        help='overwrite existing files')
    parser.add_argument('--flags', action='store_true',
                        default=True,
                        help='Add edit flags to output XML')
    parser.add_argument('--debug', action='store_true',
                        default=False,
                        help='write debug info in XML files')

    args = parser.parse_args()
    musicpath = args.musicpath
    if not musicpath:
        if 'MUSICPATH' in os.environ:
            musicpath = os.environ['MUSICPATH']
        else:
            print("\nFAILURE: Can't find a source path for music files.\n")
            parser.print_help()
            sys.exit(2)
    stats = Stats()
    all_files = get_files(musicpath, stats)
    data = scan_files(all_files)
    library = Library()
    organizer = Organizer(library)
    if args.debug:
        show_debug_data(data)
    for entry in data:
        organizer.examine_track(entry)
    print(f"\nData count: {len(data)}")
    print(f"Track count: {organizer.track_count}")
    print(f"Artist count: {len(organizer.artists)}\n")
    organizer.build_albums()
    organizer.analyze()
    organizer.push_stats(stats)
    for alb in organizer.albums:
        alb.report()

    if args.outdir:
        report_xml_files(args, organizer)

    if args.write and args.outdir:
        write_xml_files(organizer, args, stats)
    else:
        print("\nXML output was not specified.")

    stats.close()

    print(stats.report())



#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
Data objects representing musical artists.
'''

# pylint: disable=too-few-public-methods

from musicscan.data.stringtools import sanitize_for_xml


class AbstractAlbumArtist():
    '''
    Abtract class to represent an album artist.
    '''
    def __init__(self):
        self.name = None

    def __str__(self):
        return str(self.name)


class AlbumArtist(AbstractAlbumArtist):
    '''
    Class to represent a single album artist (either band or individual)
    '''
    def __init__(self, in_name):
        super().__init__()
        self.name = in_name

    def to_xml(self):
        '''
        Return the value as an XML chunk.
        '''
        return f"<artist><unkn>{sanitize_for_xml(self.name)}</unkn></artist>"


class AlbumVariousArtists(AbstractAlbumArtist):
    '''
    Class to present a designation of "various artists" on an album.
    '''
    def __init__(self):
        super().__init__()
        self.name = 'Various Artists'

    def to_xml(self):
        '''
        Return the value as an XML chunk.
        '''
        return "<variousArtists/>"



#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
Flags represent additional data added to XML output for the purpose
of identifying likely data that will probably need closer scrutiny.

Each album and song has a flag object, which will contain zero or
more flags that will be output as an XML comment with the XML elements.
'''

# pylint: disable=R0903


class Flags():
    '''
    Flags are warnings that are output in the XML
    so users can identify possible issues with
    the interpreted data.
    '''
    def __init__(self):
        self.flags = []

    def add_flag(self, in_code):
        '''
        Add a FlagCode to the array.
        '''
        self.flags.append(in_code)

    def to_xml_comment(self):
        '''
        Convert each flag code to an XML comment string value.
        '''
        if len(self.flags) > 0:
            flag_string = ''
            for f_code in self.flags:
                flag_string += FlagCodes.to_str(f_code) + ' '
            return f"<!-- EDIT FLAGS: {flag_string} -->\n"
        return ''


class FlagCodes():
    '''
    All possible flag values
    '''
    p_greatest = 1
    p_hits_compo = 2
    p_soundtrack = 3
    p_score = 4
    d_sq_brackets = 10
    d_parenthesis = 11
    l_group = 20
    p_feat_artist = 30
    p_live = 31
    p_demo = 32
    p_blank_track = 40
    p_genre_country_folk = 50
    m_year = 60

    @classmethod
    def to_str(cls, in_code):
        '''
        Convert a flag value to a usable string.
        '''
        matrix = {FlagCodes.p_greatest: 'possible_greatest_hits',
                  FlagCodes.p_hits_compo: 'possible_hits_compilation',
                  FlagCodes.p_soundtrack: 'possible_soundtrack',
                  FlagCodes.p_score: 'possible_score',
                  FlagCodes.d_sq_brackets: 'detected_square_brackets',
                  FlagCodes.d_parenthesis: 'detected_parenthesis',
                  FlagCodes.l_group: 'likely_group_artist',
                  FlagCodes.p_feat_artist: 'possible_featured_artist',
                  FlagCodes.p_live: 'possible_live_performance',
                  FlagCodes.p_demo: 'possible_demo_performance',
                  FlagCodes.p_blank_track: 'possible_blank_track',
                  FlagCodes.p_genre_country_folk:
                  'country_and_folk_genre_is_too_vague',
                  FlagCodes.m_year: 'missing_copyright_year'}
        if in_code in matrix:
            return matrix[in_code]
        return ''



#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
Data objects for music.
'''

# pylint: disable=too-many-instance-attributes

from datetime import timedelta
from musicscan.data.flags import Flags, FlagCodes
from musicscan.data.stringtools import sanitize_year


class Track():
    '''
    A track identifies a single musical track on an album,
    created through ID3 data.
    '''
    def __init__(self, in_tag):
        self.title = ''
        self.artist = ''
        self.album = ''
        self.album_o = None
        self.track_no = 0
        self.bpm = 0
        self.short_title = ''
        self.album_artist = ''
        self.composer = ''
        self.year = '0000'
        self.flags = Flags()
        self.indices = []
        self._process(in_tag)

    def _process(self, in_tag):
        self.title = in_tag.title
        self.genre = in_tag.genre
        self.artist = in_tag.artist
        self.track_no = int(in_tag.track)
        self.track_total = int(in_tag.track_total)
        if not in_tag.disc:
            self.disc_no = 1
        else:
            self.disc_no = int(in_tag.disc)
        if not in_tag.disc_total:
            self.disc_total = 1
        else:
            self.disc_total = int(in_tag.disc_total)
        self.album = in_tag.album
        if in_tag.albumartist:
            self.album_artist = in_tag.albumartist
        if in_tag.composer:
            self.composer = in_tag.composer
        # self.disc_count = int(in_tag.disc_total)
        self.duration_r = in_tag.duration
        self.duration_f = self.duration_r * 100 / int(100)
        self.duration_t = timedelta(seconds=float(self.duration_f))
        self.duration_s = Track._make_iso_interval(self.duration_t)
        if in_tag.year:
            self.year = sanitize_year(in_tag.year)
        else:
            self.flags.add_flag(FlagCodes.m_year)

    def set_album_object(self, in_album_object):
        '''
        The album object is a reference back to the object that
        contains the track.
        '''
        self.album_o = in_album_object

    @classmethod
    def _make_iso_interval(cls, in_delta):
        '''
        Generate an ISO duration value based on
        a Python timedelta object.
        '''
        string1 = str(in_delta)
        fld = string1.split(':')
        minutes = int(fld[1])
        seconds = float(fld[2])
        return f"PT{minutes}M{seconds:.2f}S"

    def get_album_artist(self):
        '''
        Return an artist value, whether it exists or not.
        '''
        return self.album_artist or self.artist

    def track_id(self):
        '''
        Return track information.
        '''
        return f"[ {self.track_no:02d} / {self.track_total:02d} " +\
               f" {self.disc_no} / {self.disc_total} ]"

    def __lt__(self, other):
        return self.track_no < other.track_no

    def __rt__(self, other):
        return self.track_no > other.track_no

    def __str__(self):
        return f"{self.track_id()} {self.title} ({self.artist})"



#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
Data objects for music.
'''

# pylint: disable=too-few-public-methods

import musicscan.data.artist
import musicscan.data.analyzer
from musicscan.data.cd import Album


class Library():
    '''
    A completed library of albums that can be iterated against.
    '''
    def __init__(self):
        self.albums = []

    def add_album(self, in_album):
        '''
        Add an album object to the library.
        '''
        self.albums.append(in_album)


class Organizer():
    '''
    A class for taking parsed id3 tags and assigning them to class
    objects for the purpose of building a library.

    Dictionary artists uses artist name string as key.
    Value is a nested dictionary based on the album name.
    That dictionary value is an array holding all the TinyTag
    objects.

    Yes, it's ugly.
    '''
    def __init__(self, in_library):
        self.track_count = 0
        self.library = in_library
        self.albums = []
        self.artists = {}

    def examine_track(self, in_tag):
        '''
        Examine a TinyTag and identify both album and artist.

        Build a data structure to accomodate the tag objects.
        '''
        artist = in_tag.albumartist or in_tag.artist
        album = in_tag.album
        if artist not in self.artists:
            self.artists[artist] = {}
            self.artists[artist][album] = [in_tag]
            self.track_count += 1
        else:
            if album not in self.artists[artist]:
                self.artists[artist][album] = [in_tag]
                self.track_count += 1
            else:
                self.artists[artist][album].append(in_tag)
                self.track_count += 1

    def build_albums(self):
        '''
        Build Album and Artist objects based on the collected
        tag data.
        '''
        for art in self.artists:
            if art == 'Various Artists':
                art_o = musicscan.data.artist.AlbumVariousArtists()
            else:
                art_o = musicscan.data.artist.AlbumArtist(art)
            for album in self.artists[art]:
                album_o = Album(art_o, album)
                for tag in self.artists[art][album]:
                    album_o.import_tag(tag)
                self.albums.append(album_o)
        for album_o in self.albums:
            album_o.finalize()

    def analyze(self):
        '''
        Perform additional tests on the data
        to flag possible inconsistencies or
        inaccuracies that can be fixed.
        '''
        analyzer = musicscan.data.analyzer.Analyzer()
        for alb in self.albums:
            analyzer.main_test(alb)

    def push_stats(self, in_stats):
        '''
        Take a Stats() object and pass data to it.
        '''
        in_stats.album_count = len(self.albums)
        in_stats.track_count = 0
        for alb in self.albums:
            for dsc in sorted(alb.discs):
                in_stats.track_count += len(alb.discs[dsc].tracks)



#!/usr/bin/env/python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
String functions.
'''

import re
import string

UNICODE_SQ = "\u2018\u2019\u201c\u201d"


def transform_string(in_value):
    '''Low level change to remove all punctuation from a string.'''
    if in_value is not None:
        no_punctuation = drop_punct(in_value)
        return no_punctuation.casefold()
    return "SHOULDNT BE HERE"


def build_filename_string(in_value):
    '''Convert all whitespace into underscores suitable for a filename'''
    level1 = transform_string(in_value)
    level2 = level1.translate(level1.maketrans(" \t\n\r", "____"))
    return level2.replace("__", "_")


def build_complete_filename(in_album, in_suffix='audiocd'):
    '''
    Build the proper filename for an output file.
    '''
    art = build_filename_string(in_album.artist.name)
    alb = build_filename_string(in_album.title)
    year = in_album.first_track().year
    name = f"{art}_{alb}-{year}-{in_suffix}.xml"
    return name


def sanitize_year(in_year):
    '''
    Ensure a year value either matches the correct pattern,
    or a year value can be extracted.
    '''
    year_p = re.compile(r'\d\d\d\d')
    mtch = year_p.match(in_year)
    return mtch.group(0) or "UNKN"


def sanitize_for_xml(in_string):
    '''
    Fix a string so it is suitable to go into
    XML output.
    '''
    out = in_string.replace('&', '&amp;') \
        .replace('<', '&lt;').replace('>', '&gt;')
    return out


def transform_ampersand(in_value):
    '''
    Simple function to transform an ampersand.
    '''
    return in_value.replace("&", "&amp;")


def drop_punct(in_str):
    '''
    Remove all punctuation symbols from a string.

    Also make a pass at removing unicode smart quotes
    '''
    level1 = in_str.translate(in_str.maketrans("", "", string.punctuation))
    out_str = level1.translate(level1.maketrans("", "", UNICODE_SQ))
    return out_str



#!/usr/bin/env python

#
# Copyright 2024 Chris Josephes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

'''
Objects to generate unique identification values for
track titles.
'''

# pylint: disable=too-few-public-methods

from musicscan.data.stringtools import drop_punct


class ShortTitleIndex():
    '''
    An album level index to track the title
    of every track for the purpose of
    maintaining a unique XML id value for the index.
    '''
    def __init__(self):
        self.t_list = {}

    def add(self, new_object):
        '''
        Try and add the short title to the index.
        If the new value conflicts, force the
        passed object to increment its index value.
        '''
        short_title_s = str(new_object)
        while short_title_s in self.t_list:
            new_object.increment()
            short_title_s = str(new_object)
        self.t_list[short_title_s] = new_object


class ShortTitle():
    '''
    A short string based on the first letter
    of the first three words of the title of the
    track, and a numerical index.  The value is
    used as a unique indentity value for
    mapping between the song and
    the physical media index.

    Ex: The song "Rock Me Amadeus" would
    have an index value of "rma01".

    The song "99 Luftballons" would
    have an index value of "n9l01"
    because the index value cannot start
    with a number.

    The song "2112" would have an identity
    string value of "n2-01" because the
    code adds a dash between title portion
    and the incrementor.

    The song "A Little Respect" would
    have an index value of "lr01" because
    the code skips the first word if it's
    an article, like 'A', 'An', or 'The'.

    If the there are two songs that have
    similar titles the incrementor value is
    automatically increased to guarauntee
    a unique value.
    '''
    def __init__(self, in_title, in_index):
        self.text = ''
        self.index = 1
        self._process(in_title, in_index)

    def _process(self, in_title, in_index):
        '''
        Check the index to ensure it has
        a unique value against other tracks.
        '''
        self.text = self._build_str(in_title)
        in_index.add(self)

    def _build_str(self, in_title):
        '''
        Ensure a short string is properly
        created based on the title of the
        track.
        '''
        lc_t = in_title.upper()
        words = lc_t.split()
        final_word = ''
        if words[0] in ['A', 'An', 'The']:
            if len(words) > 1:
                words.pop(0)
        for e_wrd in words[0:3]:
            no_punc = drop_punct(e_wrd)
            if len(no_punc) > 0:
                final_word += no_punc[0].lower()
        if final_word[0].isdigit():
            final_word = 'n' + final_word[:-1]
        return final_word

    def increment(self):
        '''
        Increment the index value.
        '''
        self.index += 1

    def __str__(self):
        if self.text[-1].isdigit():
            return f"{self.text}-{self.index:02d}"
        return f"{self.text}{self.index:02d}"
